/* Copyright 2020, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "widget/block_grid.hpp"
#include "memcard/memorycard.hpp"
#include "memcard/block.hpp"
#include "memcard/icon_fetch.hpp"
#include <cassert>
#include <algorithm>
#include <nana/gui/detail/widget_content_measurer_interface.hpp>

namespace duck::widget {
namespace {
	const constexpr int g_icon_spacing = 3;
} //unnamed namespace

namespace {
	int icon_at_coords (nana::point pos, int count, int columns, int w, int h, int spacing) {
		const int& x = pos.x;
		const int& y = pos.y;
		const int icon_rows = (count + (columns - 1)) / columns;
		assert(icon_rows > 0);

		if (x < spacing or
			y < spacing or
			x >= spacing * columns + columns * w or
			y >= spacing * icon_rows + icon_rows * h
		) {
			return count;
		}

		const int x_icon = (x - spacing) / (w + spacing);
		const int y_icon = (y - spacing) / (h + spacing);

		if ((x_icon + 1) * (w + spacing) <= x or (y_icon + 1) * (h + spacing) - spacing <= y) {
			return count;
		}

		return x_icon + y_icon * columns;
	}
} //unnamed namespace

namespace drawerbase::block_grid {
class Trigger::Measurer : public nana::dev::widget_content_measurer_interface {
public:
	Measurer(Trigger* t)
		: m_trigger{ t }
	{}

	std::optional<nana::size> measure(graph_reference graph, unsigned limit_pixels, bool /*limit_width*/) const override {
		return nana::size{51 * 3, 51 * 5};
	}

	nana::size extension() const override {
		return {g_icon_spacing, g_icon_spacing};
	}
private:
	Trigger* m_trigger;
};

Trigger::Trigger() {
	m_measurer = std::make_unique<Measurer>(this);
}

Trigger::~Trigger() noexcept = default;

void Trigger::attached (
	nana::widget& widget,
	nana::paint::graphics& graph
) {
	m_graph = &graph;
	m_widget = &widget;

	nana::API::dev::enable_space_click(widget, true);
	nana::window wd = widget;
	nana::API::tabstop(wd);
	nana::API::dev::set_measurer(widget, m_measurer.get());
}

void Trigger::detached() {
	m_widget = nullptr;
	m_graph = nullptr;
}

void Trigger::refresh (nana::paint::graphics& graph) {
	assert(m_widget);

	//draw background
	//graph.rectangle(true, nana::colors::black);
	if (not nana::API::dev::copy_transparent_background(*m_widget, graph))
		graph.rectangle(true, m_widget->bgcolor());

	if (m_highlight_index < m_block_count) {
		nana::rectangle r(
			m_highlight_index % m_columns * (m_icon_w + g_icon_spacing),
			m_highlight_index / m_columns * (m_icon_h + g_icon_spacing),
			m_icon_w + 2 * g_icon_spacing,
			m_icon_h + 2 * g_icon_spacing
		);
		graph.gradual_rectangle(
			r,
			nana::color(221, 215, 36),
			nana::color(29, 22, 12),
			//nana::color(246, 233, 200),
			true
		);
	}
}

void Trigger::mouse_down(nana::paint::graphics&, const nana::arg_mouse& mouse) {
}

void Trigger::mouse_enter (nana::paint::graphics& graph, const nana::arg_mouse& mouse) {
	m_highlight_index = static_cast<std::size_t>(icon_at_coords(
		mouse.pos,
		static_cast<int>(m_block_count),
		m_columns,
		m_icon_w,
		m_icon_h,
		g_icon_spacing
	));
	//std::cout << "mouse enter, selected " << m_highlight_index << '\n';
	refresh(graph);
}

void Trigger::mouse_leave (nana::paint::graphics& graph, const nana::arg_mouse& mouse) {
	m_highlight_index = m_block_count;
	//std::cout << "mouse leaving\n";
	refresh(graph);
}

void Trigger::mouse_move (nana::paint::graphics& graph, const nana::arg_mouse& mouse) {
	const auto new_index = static_cast<std::size_t>(icon_at_coords(
		mouse.pos,
		static_cast<int>(m_block_count),
		m_columns,
		m_icon_w,
		m_icon_h,
		g_icon_spacing
	));
	if (new_index != m_highlight_index) {
		m_highlight_index = new_index;
		refresh(graph);
	}
}

void Trigger::set_block_count (std::size_t sz) noexcept {
	if (m_highlight_index >= m_block_count)
		m_highlight_index = sz;
	else if (sz < m_block_count)
		m_highlight_index = std::min(m_highlight_index, sz);

	m_block_count = sz;
}

void Trigger::set_icon_size (nana::size sz) noexcept {
	m_icon_w = sz.width;
	m_icon_h = sz.height;
}

void Trigger::set_columns(int c) noexcept {
	m_columns = c;
}

int Trigger::columns() const noexcept {
	return m_columns;
}
} //namespace drawerbase::block_grid

BlockGrid::BlockGrid (
	nana::window wd,
	unsigned int columns,
	bool visible
) {
	this->create(wd, nana::rectangle{}, visible);
	get_drawer_trigger().set_columns(columns);
}

void BlockGrid::emplace_back (AnimationWithSize&& anim) {
	const int columns = get_drawer_trigger().columns();
	const int count = static_cast<int>(m_blocks.size());
	const constexpr int s = g_icon_spacing;
	get_drawer_trigger().set_icon_size(anim.size());
	const int w = anim.size().width;
	const int h = anim.size().height;
	const int x = (count % columns) * (s + w) + s;
	const int y = (count / columns) * (s + h) + s;

	m_blocks.push_back(std::move(anim));
	m_blocks.back().output(*this, nana::point(x, y));

	get_drawer_trigger().set_block_count(m_blocks.size());
}
} //namespace duck::widget
