/* Copyright 2020, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <type_traits>
#include <iterator>
#include <cassert>

namespace mc::psx {
template <bool Const, typename ProviderObj, template <bool> class WrappedType>
class PartIterator {
	friend class PartIterator<not Const, ProviderObj, WrappedType>;
	typedef typename std::conditional<Const, const ProviderObj, ProviderObj>::type* ProviderObjPtr;
public:
	typedef std::random_access_iterator_tag iterator_category;
	typedef std::size_t size_type;
	typedef std::ptrdiff_t difference_type;
	typedef WrappedType<Const> value_type;
	typedef WrappedType<Const> reference;
	typedef WrappedType<Const> pointer;

	PartIterator (ProviderObjPtr mc, size_type index);
	PartIterator (const PartIterator&) = default;
	PartIterator (PartIterator&&) = default;
	PartIterator();

	template <bool Const2>
	bool operator== (const PartIterator<Const2, ProviderObj, WrappedType>& other) const;
	template <bool Const2>
	bool operator!= (const PartIterator<Const2, ProviderObj, WrappedType>& other) const;
	template <bool Const2>
	bool operator< (const PartIterator<Const2, ProviderObj, WrappedType>& other) const;

	reference operator*();

	PartIterator& operator++();
	PartIterator operator++(int);
	PartIterator& operator--();
	PartIterator operator--(int);
	template <bool Const2>
	difference_type operator- (const PartIterator<Const2, ProviderObj, WrappedType>& other);
	PartIterator operator+ (std::ptrdiff_t other) const;
	PartIterator operator- (std::ptrdiff_t other) const;

private:
	ProviderObjPtr m_provider_obj;
	size_type m_index;
};

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType>::PartIterator() :
	m_provider_obj(nullptr),
	m_index(0)
{
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType>::PartIterator (ProviderObjPtr pobj, size_type index) :
	m_provider_obj(pobj),
	m_index(index)
{
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
template <bool Const2>
inline bool PartIterator<Const, ProviderObj, WrappedType>::operator== (const PartIterator<Const2, ProviderObj, WrappedType>& other) const {
	assert(other.m_provider_obj == this->m_provider_obj);
	return this->m_index == other.m_index;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
template <bool Const2>
inline bool PartIterator<Const, ProviderObj, WrappedType>::operator!= (const PartIterator<Const2, ProviderObj, WrappedType>& other) const {
	assert(other.m_provider_obj == this->m_provider_obj);
	return this->m_index != other.m_index;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
template <bool Const2>
inline bool PartIterator<Const, ProviderObj, WrappedType>::operator< (const PartIterator<Const2, ProviderObj, WrappedType>& other) const {
	assert(other.m_provider_obj == this->m_provider_obj);
	return this->m_index < other.m_index;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline auto PartIterator<Const, ProviderObj, WrappedType>::operator*() -> reference {
	assert(m_provider_obj);
	return (*m_provider_obj)[m_index];
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType>& PartIterator<Const, ProviderObj, WrappedType>::operator++() {
	++m_index;
	return *this;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType> PartIterator<Const, ProviderObj, WrappedType>::operator++(int) {
	auto cpy = *this;
	++(*this);
	return cpy;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType>& PartIterator<Const, ProviderObj, WrappedType>::operator--() {
	--m_index;
	return *this;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType> PartIterator<Const, ProviderObj, WrappedType>::operator--(int) {
	auto cpy = *this;
	--(*this);
	return cpy;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
template <bool Const2>
inline auto PartIterator<Const, ProviderObj, WrappedType>::operator- (const PartIterator<Const2, ProviderObj, WrappedType>& other) -> difference_type {
	return this->m_index - other.m_index;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType> PartIterator<Const, ProviderObj, WrappedType>::operator+ (std::ptrdiff_t other) const {
	auto retval = *this;
	retval.m_index += other;
	return retval;
}

template <bool Const, typename ProviderObj, template <bool> class WrappedType>
inline PartIterator<Const, ProviderObj, WrappedType> PartIterator<Const, ProviderObj, WrappedType>::operator- (std::ptrdiff_t other) const {
	auto retval = *this;
	retval.m_index -= other;
	return retval;
}
} //namespace mc::psx
